/*
    nullidentd.c

    Original code copyright 1999 Brian Young <bayoung@acm.org>
    bsidentd copysomethingortheother Y2K by Callum Lerwick <seg@perkinz.org>
    Some portability and efficiency patches by Peter Pentchev <roam@orbitel.bg>
    
    Licensed under the GPL, see the file COPYING included with this distribution.
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>

#include "version.h"

#define SESSION_TIMEOUT 20

#define MAX_RESPONSE 200
#define MAX_REQUEST 100

char *randusername()
{
	char *string;
	int i, length;
	// randomly choose a length betweem 3 and 8
	length = 3 + random() / (RAND_MAX / 6);
	if (string = malloc(length + 1), string == NULL)
		return NULL;
	for (i = 0; i < length; i++)
		// does your head hurt yet?
		string[i] = 'a' + random() / (RAND_MAX / ('z' + 1 - 'a'));
	// zero terminate!
	i++;
	string[i] = '\0';
	return string;
}


int write_response(int fd, char *response, int len)
{
	int retval;
	int byteswritten = 0;
	while (byteswritten < len)
	{
		retval = write(fd, response + byteswritten, len - byteswritten);
		if (retval <= 0)
		{
			/* error */
			return 1;
		}
		byteswritten += retval;
	}

	/* success */
	return 0;
}

char inbuf[2 * MAX_REQUEST];	/* should be enough */
int inread = 0;

int read_request(int fd, char *request, int maxlen)
{
	int n, len = 0;
	char *p;
	p = memchr(inbuf, '\n', inread);
	/* read until \n */
	while ((p == NULL) && (inread < maxlen - 1) && (inread < sizeof(inbuf)))
	{
		if (n = read(fd, inbuf + inread, sizeof(inbuf) - inread), n < 0)
		{
			/* error */
			return 1;
		}

		p = memchr(inbuf + inread, '\n', n);
		inread += n;
	}

	if (inread >= maxlen - 1)
	{
		/* reached maximum allowed input length */
		len = maxlen - 1;
	}
	else if (p == NULL)
	{
		/* reached end of input buffer, copy as much as needed */
		len = maxlen - 1;
	}
	else
	{
		/* found a newline! */
		len = p - inbuf + 1;
	}

	memcpy(request, inbuf, len);
	memcpy(inbuf, inbuf + len, inread - len);
	inread -= len;
	/* strip trailing CR/LF */
	while (len && ((request[len - 1] == '\r') || (request[len - 1] == '\n')))
		len--;
	/* null terminate */
	request[len] = '\0';
	/* success */
	return 0;
}

void session_timeout(int foo)
{
	exit(0);
}

int main(int argc, const char *argv[])
{
	char *userid;
	int infd;
	int outfd;
	int response_len;
	char response[MAX_RESPONSE];
	char request[MAX_REQUEST];
	if (getgid() == 0)
	{
		fprintf(stderr, "Group id is root, exiting.\n");
		return 1;
	}
	if (getuid() == 0)
	{
		fprintf(stderr, "User id is root, exiting.\n");
		return 1;
	}

	if (argc > 1)
	{
		fprintf(stderr,"bsidentd version %s (wazoo!)\nOriginal code copyright 1999 Brian Young <bayoung@acm.org>\nbsidentd copysomethingortheother Y2K by Callum Lerwick <seg@perkinz.org>\nSome portability and efficiency patches by Peter Pentchev <roam@orbitel.bg>\nRun me without args!\n",VERSION);
		return 1;
	}

	infd = fileno(stdin);
	outfd = fileno(stdout);

	// set the session timeout
	signal(SIGALRM, session_timeout);
	alarm(SESSION_TIMEOUT);

	srandom(time(NULL));

	for (;;)
	{
		/* read the request */
		if (read_request(infd, request, sizeof(request)))
		{
			/* error or timed out */
			break;
		}

		if (userid = randusername(), userid == NULL)
			break;
		/* format the response */
		response_len = snprintf(response, sizeof(response), "%.20s : USERID : UNIX : %.20s\r\n", request, userid);
		free(userid);

		/* send the line */
		if (write_response(outfd, response, response_len))
		{
			break;
		}
	}

	return 0;
}
